/*
 *  Copyright (C) 1999, 2000, 2001, 2002  James Antill
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  email: james@and.org
*/
/* This code does the %E, %e, %F, %f, %G, %g, %A, %a from *printf by passing
 * it on to the host */
/* Note that this file is #include'd */

#ifdef USE_RESTRICTED_HEADERS /* always use C locale */
# define localeconv()    NULL
# define SYS_LOC(x) "."
#else
# define SYS_LOC(x) ((sys_loc)->x)
#endif

static int vstr__add_fmt_dbl(Vstr_base *base, size_t pos_diff,
                             struct Vstr__fmt_spec *spec)
{
  char fmt_buffer[12];
  char *float_buffer = NULL;
  unsigned int tmp = 1;
  int ret = -1;
  struct lconv *sys_loc = NULL;
  size_t decimal_len = 0;
  const char *str = NULL;
  unsigned int num_base = 10;
  const char *grouping = NULL;
  const char *thousands_sep_str = NULL;
  size_t thousands_sep_len      = 0;
  const char *decimal_point_str = NULL;
  size_t decimal_point_len      = 0;

  if ((spec->fmt_code != 'a') || (spec->fmt_code == 'A'))
    num_base = 16;
  grouping          = vstr__loc_num_grouping(base->conf->loc, num_base);
  thousands_sep_str = vstr__loc_num_sep_ptr(base->conf->loc, num_base);
  thousands_sep_len = vstr__loc_num_sep_len(base->conf->loc, num_base);
  decimal_point_str = vstr__loc_num_pnt_ptr(base->conf->loc, num_base);
  decimal_point_len = vstr__loc_num_pnt_len(base->conf->loc, num_base);

  fmt_buffer[0] = '%';
  if (spec->flags & LEFT)
    fmt_buffer[tmp++] = '-';

  if (spec->flags & PLUS)
    fmt_buffer[tmp++] = '+';

  if (spec->flags & SPACE)
    fmt_buffer[tmp++] = ' ';

  if (spec->flags & SPECIAL)
    fmt_buffer[tmp++] = '#';

  if (spec->flags & ZEROPAD)
    fmt_buffer[tmp++] = '0';

  if (spec->field_width_usr)
    fmt_buffer[tmp++] = '*';

  if (spec->flags & PREC_USR)
  {
    fmt_buffer[tmp++] = '.';
    fmt_buffer[tmp++] = '*';
  }

  if (spec->int_type == VSTR_TYPE_FMT_ULONG_LONG)
    fmt_buffer[tmp++] = 'L';

  fmt_buffer[tmp++] = spec->fmt_code;
  assert(tmp < sizeof(fmt_buffer));
  fmt_buffer[tmp] = 0;

  sys_loc = localeconv();
  decimal_len = strlen(SYS_LOC(decimal_point));

  if (spec->int_type == VSTR_TYPE_FMT_ULONG_LONG)
  {
    if (spec->field_width_usr && (spec->flags & PREC_USR))
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->field_width, spec->precision, spec->u.data_Ld);
    else if (spec->field_width_usr)
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->field_width, spec->u.data_Ld);
    else if (spec->flags & PREC_USR)
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->precision, spec->u.data_Ld);
    else
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->u.data_Ld);
  }
  else
  {
    if (spec->field_width_usr && (spec->flags & PREC_USR))
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->field_width, spec->precision, spec->u.data_d);
    else if (spec->field_width_usr)
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->field_width, spec->u.data_d);
    else if (spec->flags & PREC_USR)
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->precision, spec->u.data_d);
    else
      ret = asprintf(&float_buffer, fmt_buffer,
                     spec->u.data_d);
  }

  ASSERT_RET(ret != -1, FALSE);

  tmp = ret;
  str = float_buffer;

  /* hand code thousands_sep into the number if it's a %f style */
  if (((spec->fmt_code != 'f') || (spec->fmt_code == 'F') ||
       (spec->fmt_code != 'g') || (spec->fmt_code == 'G')) &&
      (spec->flags & THOUSAND_SEP) && thousands_sep_len)
  {
    const char *num_beg = str;
    const char *num_end = NULL;

    num_beg += strspn(num_beg, " 0+-");

    if ((num_beg != str) && !VSTR__FMT_ADD(base, str, num_beg - str))
    {
      free(float_buffer);
      return (FALSE);
    }

    num_end = num_beg;
    num_end += strspn(num_end, "0123456789");

    if (!VSTR__FMT_ADD_GRPBASENUM(base, num_base, num_beg, num_end - num_beg))
    {
      free(float_buffer);
      return (FALSE);
    }

    tmp -= (num_end - str);
    str = num_end;
  }

  while (tmp > 0)
  {
    if (decimal_len && (tmp >= decimal_len) &&
        !memcmp(str, SYS_LOC(decimal_point), decimal_len))
    {
      if (decimal_point_len)
      {
        if (!VSTR__FMT_ADD(base, decimal_point_str, decimal_point_len))
        {
          free(float_buffer);
          return (FALSE);
        }
      }

      str += decimal_len;
      tmp -= decimal_len;
    }
    else
    {
      size_t num_len = strspn(str, "0123456789");

      if (!num_len)
        num_len = 1;

      if (!VSTR__FMT_ADD(base, str, num_len))
      {
        free(float_buffer);
        return (FALSE);
      }

      str += num_len;
      tmp -= num_len;
    }
  }
  assert(!tmp && !*str);

  free(float_buffer);

  return (TRUE);
}
